# color definitions
foreground      = 'white'
background      = 'black'
fieldbackground = 'grey30'
focusbackground = 'grey'

selected        = 'green2'

readonly        = 'grey10'
disabled        = 'grey40'

entry_grey      = 'grey45'
entry_white     = 'grey92'

tab             = 'grey'
